/*
 *   mp3guessenc is an mp3 encoder-guesser
 *   Copyright (C) 2002-2010 Naoki Shibata
 *   Copyright (C) 2011-2012 Elio Blanca <eblanca76@users.sourceforge.net>
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

/* mp3guessenc version 0.2 by Naoki Shibata */
/* This program is still under development. */


#define VER_PKG_NAME   "mp3guessenc"
#define VER_MAJ        0
#define VER_MIN        25
#define VER_PATCH_LEV  23
#define REL            1  /* release type 0:stable, 1:alpha, anything else means beta */
#define CODENAME       "Sparrow's Nightmare"

/*
 * release type is provided via RELEASE_STABLE/RELEASE_ALPHA/RELEASE_BETA macros
 *
 * So please, do not use REL for release check! 
 * Again, check which is true among RELEASE_STABLE/RELEASE_ALPHA/RELEASE_BETA
 */

#ifdef REL
#define RELEASE_STABLE (REL==0)
#define RELEASE_ALPHA  (REL==1)
#define RELEASE_BETA   ((REL!=0)&&(REL!=1))
#else
#define RELEASE_STABLE 0
#define RELEASE_ALPHA  1
#define RELEASE_BETA   0
#endif

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

#include "mp3guessenc.h"
#include "decode.h"
#include "scrambled.h"
#define CMD_CODENAME_LENGTH 80 /* max length of a string provided via command line */

//////////////////////////////////////////////
// Utility routines

int ExtractI4(char *buf)
{
    int x;
    x = (unsigned char)buf[0];
    x <<= 8;
    x |= (unsigned char)buf[1];
    x <<= 8;
    x |= (unsigned char)buf[2];
    x <<= 8;
    x |= (unsigned char)buf[3];
    return x;
}

short ExtractI2(char *buf)
{
    short x;
    x = (unsigned char)buf[0];
    x <<= 8;
    x |= (unsigned char)buf[1];
    return x;
}

//#define GET_DELAY

#ifdef GET_DELAY
float ExtractFP16(short i)
{
/* extracting half-precision float from a 16-bit word, as described in IEEE 754
   see also http://en.wikipedia.org/wiki/Half_precision_floating-point_format */

  /* this array saves me from writing an on-purpose routine just to calculate 2^(exponent) */
  /* and avoids math.h (and libm) dependence, which turns out in a lighter executable */
  /* it is arranged as follows: power2[0]=2^(-14)...power2[29]=2^15 */
  float power2[30]={0.00006103515625,0.0001220703125,0.000244140625,0.00048828125,0.0009765625,0.001953125,0.00390625,0.0078125,0.015625,0.03125,0.0625,0.125,0.25,0.5,1.0,2.0,4.0,8.0,16.0,32.0,64.0,128.0,256.0,512.0,1024.0,2048.0,4096.0,8192.0,16384.0,32768.0};

  unsigned char exponent=(i>>10)&0x1f;

  if (exponent==31) return 100000.0;  /* this is set to indicate inf */

  float sign;
  if (i&0x8000) sign=-1.0;
  else sign=1.0;

  short significandbits=i&0x3ff;
  float part=(float)significandbits/1024.0;

  if (exponent==0)
    return sign*power2[exponent]*part;
  else
    return sign*power2[exponent-1]*(1.0+part);
}
#endif

/////////////////////////////////////////////
// Mpeg header utilities

char head_check(unsigned long head)
/*
 * Integrity checks for an mpeg header (4 bytes)
 * Return values:
 *  0: invalid header
 *  1: valid header
 * -1: free format header
 */
{
  unsigned char ly=(head&HEADER_FIELD_LAYER)>>17;            /* layer index */
  unsigned char bi=(head&HEADER_FIELD_BITRATE)>>12;          /* bitrate index */
  unsigned char mp=(head&HEADER_FIELD_MPEG_ID)>>19;          /* mpeg version */
  unsigned char md;                                          /* audio mode */

  if ((head&HEADER_FIELD_SYNC)!=HEADER_FIELD_SYNC) return 0; /* check for sync bits */
  if (mp == 1) return 0;                                     /* check for reserved value in mpeg version field */
  if (!ly) return 0;                                         /* check for reserved value in layer field */
  if (bi == 15) return 0;                                    /* check for invalid bitrate index */
  if (((head&HEADER_FIELD_SAMPRATE)>>10) == 3 ) return 0;    /* check for reserved value in sample rate field */
  if ((head&HEADER_FIELD_EMPHASIS) == 2) return 0;           /* check for reserved value in emphasis field */

  /* the header is valid and the mpeg frame is supported! */

  if ((ly != 2)||(mp != 3)) return 1;
  /* if mpeg 1 layerII, then check for not allowed bitrate-mode combinations */
  if (bi == 0) return 1;
  md=(head&HEADER_FIELD_CHANNELS)>>6;
  if (bi<4 && md!=3) return 0;
  if (bi==5 && md!=3) return 0;
  if (bi>10 && md==3) return 0;
  /* all went ok, this mpeg1layerII header is valid */
  return 1;
}

int get_framesize(unsigned long head)
/*
 * "Just give me an audio mpeg header,
 * I will give you in return its size"
 * Works with mpeg1, mpeg2, mpeg2.5, all layers
 */
{
  unsigned char lsf,layer,bitrate_index,srate_index,mpeg,padding,samplesfactor=1;
  int size;

  lsf           = 1-((head&HEADER_FIELD_LSF)>>19);
  layer         = 3-((head&HEADER_FIELD_LAYER)>>17);   /* this goes from 0 (layerI) up to 2 (layerIII) */
  bitrate_index = (head&HEADER_FIELD_BITRATE)>>12;
  srate_index   = (head&HEADER_FIELD_SAMPRATE)>>10;
  mpeg          = (head&HEADER_FIELD_MPEG_ID)>>19;
  padding       = (head&HEADER_FIELD_PADDING)>>9;

/*
 * This `samplesfactor' is used to compensate for rounding errors when managing integers only. The
 * size formula is right just without this (aside from the pad*4) but it keeps not returning
 * multiple-of-4 values. So this way it is fixed for working in any case with layers II and III.
 * For layer I, I just multiply by 4 before returning the size.
 */
  if (!layer) samplesfactor=4;

  size = samples_tab[lsf][layer]/samplesfactor*bitrate_tab[lsf][layer][bitrate_index]*125/samplerate_tab[srate_index][mpeg]+(int)padding;

  if (layer) return size;
  else return size<<2;
}

///////////////////////////////////////////////////////
// VBR TAG related routines

void reset_tag_data(VBRTAGDATA *p)
{
  unsigned char i;
  p->alert=0;
  p->tagId=NULL;
  p->tagStartsAt=0;
  p->tagEndsAt=0;
  p->frameSize=0;
  p->infoTag=TAG_NOTAG;
  p->version=0;
  p->delay=0;
  p->tocEntries=0;
  p->sizePerTocEntry=0;
  p->framesPerTocEntry=0;
  p->tocSize=0;
  p->frames=0;
  p->bytes=0;
  p->vbr_scale=0;
  for (i=0; i<LAMETAGSIZE; i++) p->lametag[i]=0;
  return;
}

void parse_vbri_tag(VBRTAGDATA *p, char *buf)
{
  int offset=4;

  p->tagId="VBRI";
  p->infoTag=TAG_VBRITAG;
  p->version=ExtractI2(buf+offset);
  offset+=2;
#ifdef GET_DELAY
  p->delay=ExtractFP16(ExtractI2(buf+offset));
#else
  p->delay=0;
#endif
  offset+=2;
  p->vbr_scale=(int)ExtractI2(buf+offset);
  offset+=2;
  p->bytes=ExtractI4(buf+offset);
  offset+=4;
  p->frames=ExtractI4(buf+offset);
  offset+=4;
  p->tocEntries=ExtractI2(buf+offset);
  offset+=2;
  //p->scale4TocEntries=ExtractI2(buf+offset);
  offset+=2;
  p->sizePerTocEntry=ExtractI2(buf+offset);
  offset+=2;
  p->framesPerTocEntry=ExtractI2(buf+offset);
  p->tocSize=(int)p->sizePerTocEntry*(int)p->tocEntries;

  p->tagEndsAt+=4+2+2+2+4+4+2+2+2+p->tocSize;

  return;
}

void parse_xing_tag(VBRTAGDATA *p, char *buf)
{
#define XING_FLAG_FRAMES     0x0001
#define XING_FLAG_BYTES      0x0002
#define XING_FLAG_TOC        0x0004
#define XING_FLAG_VBR_SCALE  0x0008

  int offset=4,flags;

  p->tagId = "Xing";
  if (buf[0]=='I') p->infoTag = TAG_LAMECBRTAG;
  else p->infoTag = TAG_XINGTAG;

  flags = ExtractI4(buf+offset);
  offset += 4;
  p->tagEndsAt+=4;

  if (flags & XING_FLAG_FRAMES)
  {
    p->frames = ExtractI4(buf+offset);
    offset += 4;
    p->tagEndsAt+=4;
  }

  if (flags & XING_FLAG_BYTES)

  {
    p->bytes = ExtractI4(buf+offset);
    offset += 4;
    p->tagEndsAt+=4;
  }

  if (flags & XING_FLAG_TOC)
  {
    /*for(i=0;i<NUMTOCENTRIES;i++)
        store_the_toc();*/
    p->tocEntries = 100;
    p->sizePerTocEntry = 1;
    p->tocSize = 100;
    offset+=100;
    p->tagEndsAt+=100;
  }

  if (flags & XING_FLAG_VBR_SCALE)
  {
    p->vbr_scale = ExtractI4(buf+offset);
    offset+=4;
    p->tagEndsAt+=4;
  }
/*
 * Starting with lame-3.99 alpha releases the devs changed the lame tag signature so it didn't
 * start with the string `LAME', instead it just contained the capital `L' at the beginning and the
 * following expected characters in the form `Lx.xxyzz', where x is a digit and z is an optional digit.
 * Anyway the old signature `LAMEx.xxy' (y may be one of a/b/r) was restored with lame 3.99.2 release
 * in order to keep lame tag readable by old sw/hw players.
 */
  if (
  (buf[offset] == 'L' && buf[offset+1] == 'A' && buf[offset+2] == 'M' && buf[offset+3] == 'E')
  ||
  (buf[offset] == 'L' && isdigit(buf[offset+1]) && buf[offset+2] == '.' && isdigit(buf[offset+3]) && isdigit(buf[offset+4]))
  )
  {
    unsigned char i;
    for(i=0; i<LAMETAGSIZE; i++)
      p->lametag[i] = buf[offset+i];
    p->tagEndsAt+=LAMETAGSIZE;
  }

  return;
}

char GetInfoTag(VBRTAGDATA *pTagData, char *buf, int b2r, int br, int pos)
{
/*
 * Here we detect the info vbr/cbr tag (if any)
 * Return values:
 *  0: no tag
 *  1: tag found (details into pTagData)
 */
  int head;

  reset_tag_data(pTagData);

  head=ExtractI4(buf);    /* this mpeg header was already checked before calling GetInfoTag */

  if ((buf[36]=='V')&&(buf[37]=='B')&&(buf[38]=='R')&&(buf[39]=='I'))
  {
    pTagData->tagEndsAt=40;     /* of course this value will be updated into `parse_vbri_tag' */
    parse_vbri_tag(pTagData,buf+36);
//    return 1;
  }
  else
  {
    unsigned char mono,lsf;
    int offset;
    lsf  = 1-((head&HEADER_FIELD_LSF)>>19);
    mono =((head&HEADER_FIELD_CHANNELS)>>6)/3;
    offset = 4+(int)sideinfo_tab[mono][lsf];

    if (
    ((buf[offset] == 'X') && (buf[offset+1] == 'i') && (buf[offset+2] == 'n') && (buf[offset+3] == 'g'))
    ||
    ((buf[offset] == 'I') && (buf[offset+1] == 'n') && (buf[offset+2] == 'f') && (buf[offset+3] == 'o'))
    )
    {
      pTagData->tagEndsAt=offset+4;     /* of course this value will be updated into `parse_xing_tag' */
      parse_xing_tag(pTagData,buf+offset);
//      return 1;
    }
  }

  if (pTagData->infoTag) // was any tag found?
  {
    pTagData->alert=(b2r!=br);  /* if bytes2read is different from bytes actually read, then set alert for parsing */
    pTagData->frameSize=b2r;    /* save the frame size */
    pTagData->tagStartsAt=pos;  /* offset of the mpeg frame containing this tag */
    return 1;
  }
  return 0;
}

int extract_lame_string(char *lame_string,char *p,int len)
/*
 * here the search for lame extended version string is performed
 * lame stores this string into ancillary data since version 3.80b (may 6th, 2000)
 * Return values:
 * - length of lame string extracted (0 if not found)
 * - lame_string contains lame extended version string (it is leaved untouched if not found)
 * Note: this routine may be called several times since ancillary data may be found in
 * several frames, expecially in CBR files
 * it overwrites several lame strings in order to fullfill `lame_string' with data from
 * the longest of them
 */
{
//#define SHOW_LAME_STRINGS
  int i,len_before;
  char *q;

  q = strstr(p,"LAME");
  if (q == NULL)
  {
#ifdef SHOW_LAME_STRINGS
    printf("No lame string.\n");
#endif
    return 0;
  }
  len -= (q-p);

  if (lame_string[0] == '\0')
    strcpy(lame_string,"LAME"); /* this is likely to be the first call to extract_lame_string */
  else
  {
    if (lame_string[0]!='L' || lame_string[1]!='A' || lame_string[2]!='M' || lame_string[3]!='E')
    { /* lame_string is only modified here, it MUST be "" (null string) the first time,
         or it MUST start with "LAME", in any other case I want to be warned */
      printf("WARNING: buffer overwrite on lame_string! RESETTING!\n");
      for (i=0; i<LAME_STRING_LENGTH; i++) lame_string[i]=0;
      strcpy(lame_string,"LAME");
    }
  }

  i=4;   /* I'm sure q points to `LAME' so the cycle can be shortened */
  len_before=strlen(lame_string);

  while (i<len && q[i]!=0x55 && (unsigned char)q[i]!=0xAA && i<LAME_STRING_LENGTH)
  {
    lame_string[i]=q[i];
    i++;
  }

  if (i>=len_before) lame_string[i]=0;
#ifdef SHOW_LAME_STRINGS
  printf("lame_string=%s\n",lame_string);
#endif
  return i;
}

void show_info_tag (VBRTAGDATA p)
/*
 * Here we show the technical details found into the VBRI/XING tag
 * If a lame tag was found, then it's showed as well.
 */
{
  unsigned char lameTag=(p.lametag[0]!=0);

  if (p.alert) printf("Found an mpeg tag but an error occurred. Details may be inaccurate.\n");
  printf("%s tag detected into the first frame (found at %d, %d bytes long).\n",p.tagId,p.tagStartsAt,p.frameSize);
  if (p.infoTag==TAG_VBRITAG) printf("  Tag version      : %d\n",p.version);
#ifdef GET_DELAY
  if (p.infoTag==TAG_VBRITAG) printf("  Delay            : %-1.5fms\n",p.delay*1000); /* I don't trust this field as float16, since VBRI has been poorly documented... */
#endif
  printf("  File length      : %u\n",p.bytes);
  printf("  Number of frames : %u\n",p.frames);
  printf("  Quality          : %u",p.vbr_scale);
  if ((lameTag)&&(p.infoTag!=TAG_LAMECBRTAG))
  { /* sure the file is vbr encoded by lame, I just need to know whether it's abr or vbr */
    unsigned char lamemode=p.lametag[9]&15;
    if ((lamemode>2)&&(lamemode<7))
      /* it's vbr, so let's print encoder options */
      printf(" (-q %d -V %d)",(100-p.vbr_scale)%10,(100-p.vbr_scale)/10);
  }
  printf("\n  TOC              : ");
  if (p.tocSize) printf("%d bytes (%d entries, %d byte%s each)",p.tocSize,p.tocEntries,p.sizePerTocEntry,(p.sizePerTocEntry==1)?"":"s");
  else printf("no");
  printf("\n  Lame tag         : ");
  if (lameTag)
  {
    unsigned char c,i;
    printf("yes\n");
    /* print details from lame tag */
    printf("Lame tag details...\n  Lame short string  : ");
    for(i=0;i<9;i++)
      if (isprint(p.lametag[i])) putchar(p.lametag[i]); else putchar(' ');
    c=p.lametag[9]>>4;
    printf("\n  Tag revision       : ");
    if (c!=15) printf("%d",c);
    else printf("invalid!");
    c=p.lametag[9]&15;
    printf("\n  Bitrate strategy   : ");
    if ((c==2)||(c==9))
      /* it's abr */
      printf("ABR, ");
    else
    {
      if ((c==1)||(c==8))
        /* it's cbr */
        printf("CBR, ");
      else
        if ((c>2)&&(c<7))
          /* it's vbr */
          printf("VBR method %s, min ",(c==3)?"old/rh":(c==4)?"mtrh":/*if lame, c is 5*/"mt");
    }
    c=p.lametag[20];
    if (c)
    {
      printf("%d kbps",c);
      if (c==255) printf (" or higher");
    }
    else
      printf("unknown");
    c=p.lametag[10];
    printf("\n  Lowpass value      : ");
    if (c) printf("%d",c*100);
    else printf("unknown");
    c=p.lametag[19];
    printf("\n  nspsytune          : %s\n",(c&0x10)?"yes":"no");
    printf("  nssafejoint        : %s\n",(c&0x20)?"yes":"no");
    printf("  nogap continued    : %s\n",(c&0x40)?"yes":"no");
    printf("  nogap continuation : %s\n",(c&0x80)?"yes":"no");
    printf("  ATH type           : %d\n",c&0xf);
    c=p.lametag[24];
    printf("  Encoding mode      : ");
    switch (c&0x1c)
    {
      case 24:
          printf("intensity stereo");
          break;
      case 20:
          printf("auto");
          break;
      case 16:
          printf("forced MS stereo");
          break;
      case 12:
          printf("joint stereo");
          break;
      case 8:
          printf("dual channel");
          break;
      case 4:
          printf("simple LR stereo");
          break;
      case 0:
          printf("mono");
          break;
      default:
          printf("other");
    }
    printf("\n  Unwise settings    : %s\n",(c&32)?"used":"not used");
    printf("  Source frequency   : ");
    switch (c&0xc0)
    {
      case 128:
          printf("48 kHz");
          break;
      case 64:
          printf("44.1 kHz");
          break;
      case 0:
          printf("32 kHz or below");
          break;
      default:
          printf("higher than 48 kHz");
    }
    printf("\n  Originally encoded : %d bytes\n",ExtractI4(p.lametag+28));
  }
  else
    printf("no\n");
  printf("\n");
}

/*
int SeekPoint(unsigned char TOC[NUMTOCENTRIES], int file_bytes, float percent)
{
    // interpolate in TOC to get file seek point in bytes
    int a, seekpoint;
    float fa, fb, fx;

    if( percent < (float)0.0 )   percent = (float)0.0;
    if( percent > (float)100.0 ) percent = (float)100.0;

    a = (int)percent;
    if( a > 99 ) a = 99;
    fa = TOC[a];
    if( a < 99 ) {

        fb = TOC[a+1];
    } else {
        fb = (float)256.0;
    }

    fx = fa + (fb-fa)*(percent-a);

    seekpoint = (int)(((float)(1.0/256.0))*fx*file_bytes); 


    return seekpoint;
}*/

///////////////////////////////////////////////////////

int resync_mp3(FILE *fp, int *h, int pos, int *scanned)
/*
 * Scan the stream for sync bits
 * return: the offset of the next frame found
 * and the full mpeg header (stored in `h', 32 bits)
 * Further, into `scanned' we can read how many bytes
 * have been read before exiting.
 * You can give an offset into `pos', or
 * the current file position can be used setting
 * pos=RESYNC_POS_CURRENT.
 * Note: if a valid mpeg frame is found, then
 * before exiting a `fseek' is done in order to
 * restore the file pointer to the frame start
 * Note 2: despite its name, this function seeks any
 * audio mpeg frame
 */
{
  unsigned long head=0;

  if (pos==RESYNC_POS_CURRENT)
    pos=ftell(fp);
  else
  {
    if (pos<0) return -1;
    fseek(fp,pos,SEEK_SET);
  }

  head  = (unsigned char)getc(fp); head = head << 8;
  head |= (unsigned char)getc(fp); head = head << 8;
  head |= (unsigned char)getc(fp); head = head << 8;
  for(*scanned=0;*scanned<SCAN_SIZE;(*scanned)++)
  {
    head |= (unsigned char)getc(fp);
    if (feof(fp)) return -1;
    if (head_check(head))
    {
      *h=head;
      fseek(fp,pos+(*scanned),SEEK_SET);
      return pos+(*scanned);
    }
    head = head << 8;
  }

  return -1;
}

void setup_framesize(FILE *input_file, int pos, int head, streamInfo *si)
{
  int i, slot;

  if (si->lay==1) slot=4;
  else slot=1;

  if (si->freeformat)
  {
    int idontcare;
    /* I'm sure the file was already fseeked to the first frame */
    /* now I just seek to the next valid mpeg frame header */
    resync_mp3(input_file,&idontcare,pos+1,&i);
    i++; /* this is done because the search started at pos+1 */
    /* now I restore the file pointer */
    fseek(input_file,pos,SEEK_SET);

    /* ok, the first frame seems to be long `i' bytes (at most) */
    if (head&HEADER_FIELD_PADDING)
    {
      si->padded[0].frameSize=i;
      si->unpadded[0].frameSize=i-slot;
    }
    else
    {
      si->unpadded[0].frameSize=i;
      si->padded[0].frameSize=i+slot;
    }
  }
  else /* it looks like this file is using `regular' bitrate values */
  {
    int headbase=head&HEADER_ANY_BUT_BITRATE_AND_PADDING_FIELDS;
    for (i=1;i<15;i++)
    {
      si->unpadded[i].frameSize=get_framesize(headbase|(i<<12));
      si->padded[i].frameSize=si->unpadded[i].frameSize+slot;
    }
  }

}

///////////////////////////////////////////////////////

void show_id3v1(id3tag id3)
{
  unsigned char i,id3v11=(!id3.comment[28]&&id3.comment[29]);

  printf("  Title   : ");
  for(i=0;i<30;i++)
    if (isprint(id3.title[i])) putchar(id3.title[i]); else putchar('.');

  printf("\n  Artist  : ");
  for(i=0;i<30;i++)
    if (isprint(id3.artist[i])) putchar(id3.artist[i]); else putchar('.');

  printf("\n  Album   : ");
  for(i=0;i<30;i++)
    if (isprint(id3.album[i])) putchar(id3.album[i]); else putchar('.');

  printf("\n  Year    : ");
  for(i=0;i<4;i++)
    if (isprint(id3.year[i])) putchar(id3.year[i]); else putchar('.');

  printf("\n  Comment : ");
  if (id3v11)
  {
    for(i=0;i<28;i++)
      if (isprint(id3.comment[i])) putchar(id3.comment[i]); else putchar('.');
    printf("\n  Track # : %u", id3.comment[29]);
  }
  else
    for(i=0;i<30;i++)
      if (isprint(id3.comment[i])) putchar(id3.comment[i]); else putchar('.');

  printf("\n  Genre   : ");
  if (id3.genre > genre_last)
    printf("unknown");
  else
    printf("%s",genre_list[id3.genre]);
  printf("\n\n");

  return;
}

int checkid3v1(FILE *fi, id3tag *id3)
/*
 * Here the info tag id3v1.x is detected
 * This function returns either the offset of the tag
 * or the offset of EOF if no tag is found,
 * so the main cycle can stop scanning when this value has been reached
 * Note: the file pointer IS modified and then it IS restored
 */
{
  int filepos=ftell(fi),pos,read;
  fseek(fi,0,SEEK_END);
  pos = ftell(fi);

  fseek(fi,-128,SEEK_END);
  read = fread(id3,sizeof(id3tag),1,fi);
  if (read==1)
  {
    if (id3->tag[0] == 'T' && id3->tag[1] == 'A' && id3->tag[2] == 'G')
      pos -= 128;
  }
  fseek(fi,filepos,SEEK_SET);
  return pos;
}

int checkid3v2(FILE *fi, unsigned char *idMaj, unsigned char *idmin)
{
/*
 * Here the info tag id3v2 is detected
 * This function returns the size of the tag (if detected)
 * so the main cycle can choose if either show it or skip it.
 * Also, idMaj and idmin provide id3V2 major and minor version number.
 * If the tag is not found, return value is 0.
 * If errors are detected, return value is -1.
 * Note: this routine DOES read data from the input file
 * but it DOES restore the file position before exiting.
 */

  int filepos=ftell(fi),id3v2size=0;
  unsigned char c1,c2,c3,c4,c5;
  fseek(fi,0,SEEK_SET);
  c1 = getc(fi);
  c2 = getc(fi);
  c3 = getc(fi);
  c4 = getc(fi);
  c5 = getc(fi);

  if (c1 == 'I' && c2 == 'D' && c3 == '3')
  {   /* has ID3v2 */
    *idMaj=c4; *idmin=c5;
    c5 = getc(fi);   /* get tag flags */
    c1 = getc(fi);
    c2 = getc(fi);
    c3 = getc(fi);
    c4 = getc(fi);
    if (!(c1&0x80) && !(c2&0x80) && !(c3&0x80) && !(c4&0x80) && !(c5&0x0f))
    {
      id3v2size = c1*2097152+c2*16384+c3*128+c4 +10; /* 10 is due to tag header (10 bytes long) */
      if (c5&0x10) id3v2size+=10;  /* is tag footer present? */
    }
    else
      id3v2size=-1;
  }

  fseek(fi,filepos,SEEK_SET);
  return id3v2size;
}

///////////////////////////////////////////////////////

void initBits(FILEBITS *fb,FILE *fp, int eobs)
{
  fb->fp = fp;
  fb->nLastBits = 0;
  fb->location = 8*ftell(fp);
  fb->bit_size = eobs*8;  /* end_of_bitstream is already known, it's located at id3pos */
/*  fseek(fp,0,SEEK_END);
  fb->bit_size = ftell(fp)*8;
  fseek(fp,fb->location/8,SEEK_SET); */
}

int ftellBits(FILEBITS *fb)
{
  return fb->location;
}

void fseekBits(FILEBITS *fb,int offset,int whence)
{
  switch(whence)
    {
    case SEEK_SET:
      fb->location = offset;
      break;
    case SEEK_CUR:
      fb->location += offset;
      break;
    case SEEK_END:
      fb->location = fb->bit_size;
      if (offset<0) fb->location+=offset; /* cannot move beyond EObitstream */
      /* fseek(fb->fp,0,SEEK_END);
      fb->location = 8*ftell(fb->fp)+offset; */
      break;
    }
  fseek(fb->fp,fb->location/8,SEEK_SET);
/*
 * I choose not to check for EOF here, since fseeking at EOF doesn't set eof() on `true'.
 * Next calls to getc will, and subsequent checks for eof() will be able to detect the failure.
 */

  /* Ok, now fill the register lastBits and prepare for next readBits calls */
  fb->lastBits = (unsigned char)getc(fb->fp);
  fb->lastBits = fb->lastBits<<8;
  fb->lastBits |= (unsigned char)getc(fb->fp);
  fb->lastBits = fb->lastBits<<8;
  fb->lastBits |= (unsigned char)getc(fb->fp);
  fb->lastBits = fb->lastBits<<8;
  fb->lastBits |= (unsigned char)getc(fb->fp);
  
  fb->nLastBits = 32 - fb->location % 8;
}

unsigned long int readBits(FILEBITS *fb,int n)
/*
 * This routine provides a bit-oriented reading facility.
 * It stores bytes read from the stream into a buffer called lastBits
 * (it is managed as a scrolling register, the output is from the most significant bits)
 * and counts the valid (unused) ones into nLastBits (the remaining least significant bits).
 * User calls cause the high order bits to be extracted and returned as a result,
 * discarding them from the nLastBits count.
 * So, when the left valid bits are <= than 24, then this routine can fullfill the register
 * with (at least) a new byte.
 * Note: even if an EOF check is performed, no flag is returned about the result.
 */
{
  unsigned long int ret = 0;
  int k;
  unsigned char c;

  fb->location += n;

/*
 * I think this first `while' cycle is here since in the early versions of fseekBits
 * the lastBits buffer was not filled after a seek operation, so this check was mandatory
 * As a safe move, I will keep it here for a while and see whether the alert is printed or not
 */
  while(fb->nLastBits <= 24) {
    int tmp;
printf("*** while statement @ %d ***\n",__LINE__);
    tmp = getc(fb->fp);
    c = tmp;
    fb->lastBits = fb->lastBits << 8;
    if (tmp != EOF) fb->lastBits = fb->lastBits | c;
    fb->nLastBits += 8;
  }

  if (n <= fb->nLastBits) k = n; else k = fb->nLastBits;


  ret = (fb->lastBits >> (fb->nLastBits - k)) & ((1 << k)-1);
  fb->nLastBits -= k;
  n -= k;

  while(fb->nLastBits <= 24) {
    int tmp;
    tmp = getc(fb->fp);
    c = tmp;
    fb->lastBits = fb->lastBits << 8;
    if (tmp != EOF) fb->lastBits = fb->lastBits | c;
    fb->nLastBits += 8;
  }

  ret = ret << n;
  ret = ret | ((fb->lastBits >> (fb->nLastBits - n)) & ((1 << n)-1));
  fb->nLastBits -= n;
    
  return ret;
}

///////////////////////////////////////////////////////

/*
 * you can take 8*511 bit
 * from the bit reservoir and at most 8*1440 bit from the current
 * frame (320 kbps, 32 kHz), so 8*1951 bit is the largest possible
 * value for MPEG1 and 2)
 * Fix: the biggest supported frame is now 8*2880 bit due to a free
 * format stream (640 kbps, 32 kHz). So the largest value is 3391 bytes.
 */

#define LARGEST_BUFFER 3400   // 3400 > 3391

typedef struct PART23BUF {
  unsigned char buf[LARGEST_BUFFER];
  int pos,len;
} PART23BUF;

// In the worst case, 9 bitstream frames have to be received before decoding can start.
#define NP23B (9+3)
PART23BUF p23b[NP23B]; 
int p23b_in=0,p23b_out=0,p23b_pos=0;

int seek_p23b(int pos)
{
  int i;
  for(i=0;i<NP23B;i++)
  {
    //printf("seek_p23b : pos=%d, i=%d, p23b[i].pos=%d, p23b[i].len=%d\n",pos,i,p23b[i].pos,p23b[i].len);
    if (p23b[i].pos <= pos && pos < p23b[i].pos+p23b[i].len-(p23b[i].pos & 7))
    {
      p23b_out = i;
      p23b_pos = pos-p23b[p23b_out].pos+(p23b[p23b_out].pos & 7);
      return 1;
    }
  }

  return 0;
}

int skip_p23b(int len)
{
  while(len)
    {
      //printf("skip_p23b : len=%d, p23b_in=%d, p23b_out=%d, p23b_pos=%d, p23b[p23b_out].pos=%d, p23b[p23b_out].len=%d\n",len,p23b_in,p23b_out,p23b_pos,p23b[p23b_out].pos,p23b[p23b_out].len);
      if (p23b[p23b_out].len-p23b_pos < len)
      {
    len -= p23b[p23b_out].len-p23b_pos;
    p23b_out++;
    if (p23b_out == NP23B) p23b_out = 0;
    p23b_pos = p23b[p23b_out].pos & 7;
    if (p23b_in == p23b_out) return len;
      }
      else
      {
    p23b_pos += len;
    len = 0;
      }
    }

  return 0;
}

int skipback_p23b(int len)
{
  while(len)
    {
      //printf("skipback_p23b : len=%d, p23b_in=%d, p23b_out=%d, p23b_pos=%d, p23b[p23b_out].pos=%d, p23b[p23b_out].len=%d\n",len,p23b_in,p23b_out,p23b_pos,p23b[p23b_out].pos,p23b[p23b_out].len);
      if (p23b[p23b_out].len-(p23b[p23b_out].pos & 7) < len)
      {
    len -= p23b[p23b_out].len-(p23b[p23b_out].pos & 7);
    p23b_out--;

    if (p23b_out == -1) p23b_out = NP23B-1;
    p23b_pos = p23b[p23b_out].len-1;
    if (p23b_in == p23b_out) return len;
      }
      else
      {
    p23b_pos -= len;
    len = 0;
      }
    }

  return 0;
}

int tell_p23b(void)
{
//printf("p23b_pos=%u, p23b_out=%u, p23b[p23b_out].pos=%u, (p23b[p23b_out].pos & 7)=%u\n",p23b_pos,p23b_out,p23b[p23b_out].pos,(p23b[p23b_out].pos & 7));
  return p23b_pos + p23b[p23b_out].pos - (p23b[p23b_out].pos & 7);
}

void load_into_p23b(FILEBITS *fb,int len)
{
  unsigned int bytes2read;
  p23b[p23b_in].pos = ftellBits(fb);
  p23b[p23b_in].len = len + (p23b[p23b_in].pos & 7);
/*
 * I have to use `fseek' here instead of `fseekBits' since `fseekBits'
 * has to be used in conjunction with `readBits', not `fread'.
 */
  fseek(fb->fp,fb->location/8,SEEK_SET);
  bytes2read=len/8;
  if (len%8) bytes2read++;
  if (fread(p23b[p23b_in].buf,1,bytes2read,fb->fp)!=bytes2read) printf("Warning: read failure.\n");
  fseekBits(fb,p23b[p23b_in].pos-(p23b[p23b_in].pos&7)+p23b[p23b_in].len,SEEK_SET);
  if (p23b_in == p23b_out) p23b_pos = p23b[p23b_out].pos & 7;
  p23b_in++;
  if (p23b_in == NP23B) p23b_in = 0;
}

/*
unsigned long int readBits_p23b(int n)
{
  unsigned long int ret = 0;

  if (p23b_in == p23b_out) return 0;

  while(n) {
    if (n >= 8 && p23b[p23b_out].len-p23b_pos > 8) {
      ret = (ret << 8) |
    (((p23b[p23b_out].buf[p23b_pos>>3] << (p23b_pos & 7)) |
      (p23b[p23b_out].buf[(p23b_pos>>3)+1] >> (8-(p23b_pos & 7)))) & 0xff);
      p23b_pos += 8; n -= 8;
    } else {
      int k = p23b[p23b_out].len-p23b_pos;
      if (n < k) k = n;
      ret = (ret << k) |
    (((p23b[p23b_out].buf[p23b_pos>>3] << (p23b_pos & 7)) |
      (p23b[p23b_out].buf[(p23b_pos>>3)+1] >> (8-(p23b_pos & 7)))) & (((1 << k)-1) << (7-k)));
      p23b_pos += k; n -= k;

      if (p23b_pos == p23b[p23b_out].len) {
    p23b_out++;
    if (p23b_out == NP23B) p23b_out = 0;
    p23b_pos = p23b[p23b_out].pos & 7;
    if (p23b_in == p23b_out) return ret << n;
      }
    }
  }

  return ret;
}
*/

// read byte-aligned 8 bits
unsigned char readOneByte_p23b(void)
{
  unsigned char ret = 0;

  if (p23b_in == p23b_out) return 0;

  for(;;)
  {
    p23b_pos = (p23b_pos+7) & ~7;
    if (p23b[p23b_out].len-p23b_pos >= 8)
    {
      ret = p23b[p23b_out].buf[p23b_pos>>3];
      p23b_pos += 8;
      return ret;
    }
    else
    {
      p23b_out++;
      if (p23b_out == NP23B) p23b_out = 0;
      p23b_pos = p23b[p23b_out].pos & 7;
      if (p23b_in == p23b_out) return 0;
    }
  }

  return 0;
}

//////////////////////////////////////////////////////////
char freeformat(int head)
{
    return ((head&HEADER_FIELD_BITRATE)==0);
}

char layerIII(int head)
{
    return ((head&HEADER_FIELD_LAYER)==0x20000);
}

void scan_layerIII(streamInfo *si, FILEBITS *fb, detectionInfo *di, int head, int startOfFrame, int framesize, int *endOfLastPart3, char *lame_string)
{
    int main_data_begin,part2_3_length,part1_length,pos_main_data_begin;
    char gr,ngr=((head&HEADER_FIELD_LSF)>>19)+1,ch,nch=(((head&HEADER_FIELD_CHANNELS)>>6)==3)?1:2,scfsi_band,window_switching_flag,block_type,mixed_block_flag;

/* (si->id == 3) means mpeg version 1 */
    if (si->id==3)
        main_data_begin = readBits(fb,9);
    else
        main_data_begin = readBits(fb,8);

    if (main_data_begin > si->reservoirMax)
        si->reservoirMax = main_data_begin;

    if (si->id==3)
    { /* si->mode == 3 means single channel mode */
        if (si->mode==3)
            readBits(fb,5); // private_bits
        else
            readBits(fb,3); // private_bits
    }
    else
    {
        if (si->mode==3)
            readBits(fb,1); // private_bits
        else
            readBits(fb,2); // private_bits
    }

    if (si->id == 3)
    {
        for(ch=0;ch<nch;ch++)
            for(scfsi_band=0;scfsi_band<4;scfsi_band++)
                di->usesScfsi = readBits(fb,1) || di->usesScfsi;
    }

    part2_3_length = 0;

    for(gr=0;gr<ngr;gr++)
    {
        for(ch=0;ch<nch;ch++)
        {
            part2_3_length += readBits(fb,12);
//            readBits(&fb,9); // big_values
//            readBits(&fb,8); // global_gain
            readBits(fb,9+8); /* a single call brings time saving */
            if (si->id == 3)
                readBits(fb,4); // scalefac_compress
            else
                readBits(fb,9); // scalefac_compress
            window_switching_flag = readBits(fb,1);
            if (window_switching_flag == 1)
            {
                block_type = readBits(fb,2); // block_type
                mixed_block_flag = readBits(fb,1); // mixed_block_flag
//                for(region=0;region<2;region++) readBits(&fb,5); // table_select
                readBits(fb,5+5); /* a single call brings time saving */
//                for(window=0;window<3;window++) readBits(&fb,3); // subblock_gain
                readBits(fb,3+3+3); /* a single call brings time saving */
                if (block_type == 2)
                {
                    if (mixed_block_flag)
                        di->blockCount[2]++;
                    else
                        di->blockCount[1]++;
                }
                else
                    di->blockCount[0]++;
            }
            else
            {
//                int region;
//                for(region=0;region<3;region++) readBits(&fb,5); // table_select
//                readBits(&fb,4); // region0_count
//                readBits(&fb,3); // region1_count
                readBits(fb,5+5+5+4+3); /* a single call brings time saving */
                di->blockCount[0]++;
            }
            if (si->id == 3)
                readBits(fb,1); // preflag
            di->usesScalefacScale = readBits(fb,1) || di->usesScalefacScale; // scalefac_scale
            readBits(fb,1); // count1table_select
        }
    }

    part1_length = ftellBits(fb) - startOfFrame;

    //printf("framenum = %d, pos*8 = %d, framesize = %d, startOfFrame = %d, main_data_begin*8 = %d, part1 = %d, part2_3 = %d, endOfLastPart3 = %d\n",framenum,pos*8,framesize*8,startOfFrame,main_data_begin*8,part1_length,part2_3_length,endOfLastPart3);

    // load part 2 and part 3 into buffer

    //printf("load : %d bit from %d\n",framesize*8-part1_length,ftellBits(&fb));
    load_into_p23b(fb,framesize*8-part1_length);

    if (main_data_begin == 0)
    {
        pos_main_data_begin = startOfFrame+part1_length;
    }
    else
        if (seek_p23b(startOfFrame-1))
        {
            if (skipback_p23b(main_data_begin*8-1))
            {
                printf("frame %d : More bits in reservoir are needed to decode this frame.\n",si->totFrameNum);
                pos_main_data_begin = -1;
            }
            else
            {
                pos_main_data_begin = tell_p23b();
                //printf("endOfLastPart3=%d,mdb=%d\n",endOfLastPart3,pos_main_data_begin);
            }
        }
        else
        {
            printf("frame %d : more bits in reservoir are needed to decode this frame.\n",si->totFrameNum);
            pos_main_data_begin = -1;
        }

    if (pos_main_data_begin != -1)
    {   // decode part2, part3 and ancillary bits
        if (*endOfLastPart3 != -1 && seek_p23b(*endOfLastPart3))
        {
            seek_p23b((*endOfLastPart3+7)&~7);
            if (pos_main_data_begin-tell_p23b() >= 8)
            {
                char buf[BUFFER_SIZE],*p;
                unsigned char c;
                p = buf;
                while(pos_main_data_begin-tell_p23b() >= 8)
                {
                    c = readOneByte_p23b();
                    if ((p-buf)<BUFFER_SIZE)
                        *p = c;
                    else
                        printf("%s @ %u: *** Stack smashing prevention! ***\nBuffer overflow avoided, double-check the code!!! (pointer=%u)\n",__FILE__,__LINE__,(unsigned int)(p-buf));
                    p++;
                }
                extract_lame_string(lame_string,buf,p-buf);
            }
        }
    }
    *endOfLastPart3 = -1;

    //printf("pos_main_data_begin=%d, part2_3_length=%d\n",pos_main_data_begin,part2_3_length);
    if (seek_p23b(pos_main_data_begin))
    {
        if (skip_p23b(part2_3_length) == 0)
            *endOfLastPart3 = tell_p23b();
        else
            printf("B!\n");
    }
    else
        printf("A!\n");

    return;
}

void scan(FILE *input_file, int *pos, streamInfo *si, detectionInfo *di, int id3pos, char *lame_string)
{
    int head,i,framesize,endOfLastPart3=-1,oldpos;
    unsigned int head_pattern,mask;
    FILEBITS fb;
/*
 * I'm sure I'm about to read the first valid audio frame of the stream,
 * since `scan' is called right after a successfull resync_mp3().
 * Now resync_mp3() is called again in order to get the value of `head'
 * and to properly set the file pointer.
 */
    oldpos=*pos=resync_mp3(input_file, &head, *pos, &i);
/*
 * Gathering basic properties of the mpeg stream, from now on I will assume the other frames have
 * the same mpeg version/layer/freq/etc...
 */

    si->freeformat = (head&HEADER_FIELD_BITRATE)==0;
    si->lsf        = 1-((head&HEADER_FIELD_LSF)>>19);
    si->id         = (head&HEADER_FIELD_MPEG_ID)>>19;
    si->freq       = samplerate_tab[(head&HEADER_FIELD_SAMPRATE)>>10][si->id];
    si->lay        = 4-((head&HEADER_FIELD_LAYER)>>17);
    si->mode       = (head&HEADER_FIELD_CHANNELS)>>6; /* channel mode (stereo, Jstereo, Dchannel, mono) */

    si->crc        = (head&HEADER_FIELD_CRC) >> 16;
    si->copyright  = (head&HEADER_FIELD_COPYRIGHT) >> 3;
    si->original   = (head&HEADER_FIELD_ORIGINAL) >> 2;
    si->emphasis   = head&HEADER_FIELD_EMPHASIS;

    setup_framesize(input_file, *pos, head, si);

    for(i=0;i<15;i++)
    {
        si->bitrateCount[i]        = 0;
        si->unpadded[i].frameCount = 0;
        si->padded[i].frameCount   = 0;
    }
    for(i=0;i<5;i++) di->modeCount[i]  = 0;
    for(i=0;i<3;i++) di->blockCount[i] = 0;

    /* mask selection: looks like layer I/II when encoding jstereo, switch between different stereo
       modes (they put 0x0, 0x1 or 0x2 in the channels field), while layer III only uses
       jstereo frames (0x1 in the channels field) and puts further informations into the mode
       extension field */
    if (layerIII(head)||((head&HEADER_FIELD_CHANNELS>>6)==3))
      mask=HEADER_CONSTANT_FIELDS_STRICT;
    else
      mask=HEADER_CONSTANT_FIELDS;

    /* using a pattern will let me check a whole header in a single `if' statement */
    head_pattern = head & mask;

    initBits(&fb,input_file,id3pos);

    while (*pos!=-1)
    {
        if (*pos!=oldpos)
            printf("Resync done at frame %d, skipped %d bytes.\n",si->totFrameNum,*pos-oldpos);

        /* this call is needed in order to properly set the `fb' struct, resync was just done */
        /* the file pointer is moved right after the header */
        fseekBits(&fb,8*(*pos)+32,SEEK_SET);

        /* now check the header we've just read */
        if ((head&mask)!=head_pattern)
        {
            /* do I need a real full-featured error detection routine? mhhh... */
            printf("Sync error at frame %d, offset %d. Seeking for resync...\n",si->totFrameNum,*pos);
            *pos=resync_mp3(input_file, &head, (*pos)+1, &i);
            continue;
        }

        if (!(head&HEADER_FIELD_CRC)) readBits(&fb,16);
        i=(head&HEADER_FIELD_BITRATE)>>12;

        if (head&HEADER_FIELD_PADDING)
        {
            di->usesPadding = 1;
            framesize=si->padded[i].frameSize;
        }
        else
            framesize=si->unpadded[i].frameSize;

        /* check integrity for this frame */
        if (*pos+framesize>id3pos)
        {
            /* we cannot read the whole frame, so skip the rest */
            *pos=resync_mp3(input_file, &head, (*pos)+1, &i);
            continue;
        }

        /* if this point is reached, then I'm sure the frame
           is readable as a whole, so I assume it's valid */
        if (head&HEADER_FIELD_PADDING)
            si->padded[i].frameCount++;
        else
            si->unpadded[i].frameCount++;
        si->bitrateCount[i]++;

        if (layerIII(head))
            scan_layerIII(si,&fb,di,head,(*pos)*8,framesize,&endOfLastPart3,lame_string);

        si->totFrameLen += framesize;
        si->totFrameNum ++;

        if ((head&HEADER_FIELD_CHANNELS)>>6 == 1) /* if jstereo, then update mode_extension data */
        {
            di->modeCount[(head&HEADER_FIELD_MODEXT)>>4]++;
            /* when handling layer I/II we may find a stereo frame at the beginning, jstereo frames
               (signaling the file is encoded as jstereo) may be encountered later */
            si->mode=1;
        }
        else
            di->modeCount[4]++;

        *pos += framesize;

        oldpos=*pos;
        *pos=resync_mp3(input_file, &head, *pos, &i);

    } /* END OF `while' CYCLE */

    if (layerIII(head))
    {
        if (endOfLastPart3 != -1 && seek_p23b(endOfLastPart3))
        {
            // extract ancillary
            int mdb;
            seek_p23b(oldpos-1); /* replaced startOfFrame with oldpos */
            mdb = tell_p23b()+1;
            seek_p23b((endOfLastPart3+7)&~7);
            if (mdb-tell_p23b() > 8)
            {
                char buf[BUFFER_SIZE],*p;
                unsigned char c;
                p = buf;
                //printf("anc start pos : %d\n",(endOfLastPart3+7)&~7);
                //printf("ancillary(%d,%d) : ",framenum,((endOfLastPart3+7)&~7)/8);
                //printf("ancillary(%d) : ",framenum);
                while(mdb-tell_p23b() > 8)
                {
                    c = readOneByte_p23b();
                    if ((p-buf)<BUFFER_SIZE)
                        *p = c;
                    else
                        printf("%s @ %u: *** Stack smashing prevention! ***\nBuffer overflow avoided, double-check the code!!! (pointer=%u)\n",__FILE__,__LINE__,(unsigned int)(p-buf));
                    p++;
                }
                extract_lame_string(lame_string,buf,p-buf);
            }
        }
    }

    /* `oldpos' now stores the last `to-be-checked' byte offset, is it exactly the same as id3pos? */
    if (oldpos<id3pos)
        printf("Detected unrecognizable stuff at %d (length in bytes: %d).\n",oldpos,id3pos-oldpos);
}
/////////////////////////////////////////////////////////////////
///////////////////      END SCAN      //////////////////////////
/////////////////////////////////////////////////////////////////

/* This routine prints lots of detailed informations about the processed mpeg bitstream */
void show_stream_info(streamInfo si, detectionInfo di, char *lame_string, int framesize)
{
  int i;
  char max_len;
  double duration;

  printf("\nDetected MPEG stream version %s layer ",!(si.id)? "2.5" : (si.id==2)? "2" : "1");
  for (i=0;i<si.lay;i++) printf("I");
  printf(", details follow.\n  File size               : %u bytes\n",si.filesize);
  printf("  Audio stream size       : %u bytes",si.totFrameLen);
  /* if the file has a header tag, then show the total size */
  if (framesize) printf (" (including tag: %u)",si.totFrameLen+framesize);
  /* just a simple check */
  if (si.totFrameLen>si.filesize) printf(" (!!!)");
  duration=(double)si.totFrameNum*(double)samples_tab[si.lsf][si.lay-1]/(double)si.freq;
  printf("\n  Length                  : %u:%02u:%02u.%03.0f (%1.3f seconds)\n",((int)duration/3600),(((int)duration%3600)/60),((int)duration%60),(1000*(duration-(int)duration)),duration);
  printf("  Data rate               : %1.1f kbps",(double)si.totFrameLen*(double)si.freq*0.008/((double)si.totFrameNum*(double)samples_tab[si.lsf][si.lay-1]));
  if (si.freeformat) printf(" (free format bitstream)");
  printf("\n  Number of frames        : ");
  max_len=printf("%d",si.totFrameNum);
  printf("\n  Audio samples per frame : %d\n",samples_tab[si.lsf][si.lay-1]);
  printf("  Audio frequency         : %d Hz\n",si.freq);
  printf("  Encoding mode           : ");
  switch (si.mode)
  {
    case 3:
      printf("mono");
      break;
    case 2:
      printf("dual channel");
      break;
    case 1:
      printf("joint stereo");
      break;
    case 0:
      printf("stereo");
      break;
  }
  printf("\nFlags\n  Error protection : %s\n",si.crc ? "no":"yes");
  printf("  Copyrighted      : %s\n",si.copyright ? "yes":"no");
  printf("  Original         : %s\n",si.original ? "yes":"no");
  printf("  Emphasis         : %s\n\n",si.emphasis == 0 ? "none" : si.emphasis == 1 ? "50/15ms" : "CCITT");

  if (si.mode==1) /* Jstereo? */
  {
    double sum=(double)(di.modeCount[0]+di.modeCount[1]+di.modeCount[2]+di.modeCount[3]+di.modeCount[4]);
    if (si.lay==3)
    {
      printf("Mode extension: stereo mode frame count\n");
      if (di.modeCount[0]) printf("  simple stereo                 : %*d (%4.1f%%)\n",max_len,di.modeCount[0],(double)di.modeCount[0]*100.0/sum);
      if (di.modeCount[1]) printf("  intensity stereo              : %*d (%4.1f%%)\n",max_len,di.modeCount[1],(double)di.modeCount[1]*100.0/sum);
      if (di.modeCount[2]) printf("  mid-side stereo               : %*d (%4.1f%%)\n",max_len,di.modeCount[2],(double)di.modeCount[2]*100.0/sum);
      if (di.modeCount[3]) printf("  intensity and mid-side stereo : %*d (%4.1f%%)\n",max_len,di.modeCount[3],(double)di.modeCount[3]*100.0/sum);
      printf("  -----------------------------\n  sum                           ");
    }
    else
    {
      printf("Mode extension: intensity stereo applied in\n");
      if (di.modeCount[0]) printf("  bands  4 to 31     : %*d (%4.1f%%)\n",max_len,di.modeCount[0],(double)di.modeCount[0]*100.0/sum);
      if (di.modeCount[1]) printf("  bands  8 to 31     : %*d (%4.1f%%)\n",max_len,di.modeCount[1],(double)di.modeCount[1]*100.0/sum);
      if (di.modeCount[2]) printf("  bands 12 to 31     : %*d (%4.1f%%)\n",max_len,di.modeCount[2],(double)di.modeCount[2]*100.0/sum);
      if (di.modeCount[3]) printf("  bands 16 to 31     : %*d (%4.1f%%)\n",max_len,di.modeCount[3],(double)di.modeCount[3]*100.0/sum);
      if (di.modeCount[4]) printf("Simple stereo frames : %*d (%4.1f%%)\n",max_len,di.modeCount[4],(double)di.modeCount[4]*100.0/sum);
      printf("  ------------------\n  sum                ");
    }
    printf(": %d\n\n",(int)sum);
  }

  if (si.lay==3)
  {
    printf("long block granules  : %d\n",di.blockCount[0]);
    printf("short block granules : %d\n",di.blockCount[1]);
    printf("mixed block granules : %d\n\n",di.blockCount[2]);

    if (di.usesScalefacScale) printf("scalefac_scale is used\n");
    if (di.usesScfsi) printf("Scalefactor selection information bit is used.\n");
    printf("max reservoir : %d\n",si.reservoirMax);
  }
  if (di.usesPadding) printf("Padding is used.\n");
  
  printf("\nFrame histogram\n");
  for(i=0;i<15;i++)
    if (si.bitrateCount[i])
    {
      if (i==0)
        printf("Custom frames : ");
      else
        printf("%4d kbps : ",bitrate_tab[si.lsf][si.lay-1][i]);
      printf("%*d (%4.1f%%), size distr: [",max_len,si.bitrateCount[i],(double)si.bitrateCount[i]*100.0/(double)si.totFrameNum);

      if (si.unpadded[i].frameCount)
        printf("%*d x%4d b",max_len,si.unpadded[i].frameCount,si.unpadded[i].frameSize);

      if ((si.unpadded[i].frameCount)&&(si.padded[i].frameCount))
        printf(", ");

      if (si.padded[i].frameCount)
        printf("%*d x%4d b",max_len,si.padded[i].frameCount,si.padded[i].frameSize);

      printf("]\n");
    }

  printf("\n%d header errors.\n",si.nSyncError);

  if (lame_string[0] != '\0') printf("\nLame string : %s\n",lame_string);
}

char *guesser(detectionInfo di)
{
/* Note: Xing and helix encoders count in their xing tag also the xing tag itself
 * as a valid frame, resulting in a +1 difference with actual audio frames amount.
 * You have the same behaviour with FhG encoders and their VBRI tag
 */
#if 0
int l;
for (l=0; l<3; l++) printf("blockCount[%u]=%u ",l,di.blockCount[l]);
printf("\n");
for (l=0; l<4; l++) printf("modeCount[%u]=%u ",l,di.modeCount[l]);
printf("\nusesScfsi         : %s\nusesScalefacScale : %s\nusesPadding       : %s",di.usesScfsi?"yes":"no",di.usesScalefacScale?"yes":"no",di.usesPadding?"yes":"no");
#endif
  if (di.blockCount[1] == 0)
  {  /* helix encoder (which is based on xing source code) DOES use short blocks! */
    if (di.modeCount[1])
      return "Xing (very old)";
    else
    {
      if (di.usesScfsi)
        return "Xing (new)";
      else
        return "Xing (old)";
    }
  }
  else
    if (di.usesScfsi)
    {  /* gogo-no-coda (?) seems to be a japanese fork of lame using asm-optimizations in order to provide max encoding speed; it DOESN'T add a lame tag */
      if (di.usesScalefacScale)
        return "Lame";
      else
        return "Lame (old) or m3e";
    }
    else
      if (di.usesScalefacScale)
      {
        if (di.usesPadding)
          return "FhG (l3enc, fastenc or mp3enc)";
        else
          return "FhG (ACM or producer pro)";
      }
      else
        return "dist10 encoder or other encoder";
}

///////////////////////////////////////////////////////

void print_version(char verbose)
{
  printf ("%s-%u.%u",VER_PKG_NAME,VER_MAJ,VER_MIN);
  if (RELEASE_STABLE) printf(".%u final",VER_PATCH_LEV);
  if (RELEASE_ALPHA) printf (" alpha");
  if (RELEASE_BETA) printf (" beta");
  if (RELEASE_ALPHA || RELEASE_BETA) printf (" %u", VER_PATCH_LEV);
  if (verbose)
  {
#ifdef CODENAME
    printf(", codename \"%s\"",CODENAME);
#endif
    printf("\nExecutable built on %s.", __DATE__);
  }
  printf("\n");
}

void usage()
{
  print_version(0);
  printf("\
Show detailed informations about mpeg audio files and guess the encoder used.\n\
Usage:\n\
  -h\tprint this help screen\n\
  -V\tprint version informations\n\
\n");
}

///////////////////////////////////////////////////////

int main(int argc,char **argv)
{
  char lame_string[LAME_STRING_LENGTH], buf[LARGEST_BUFFER];
  id3tag id3;
  unsigned char id3V2Maj, id3V2min;
  int pos=0, i, option, head, bytes2read, id3pos=0, id3v2size;
  FILE *input_file;
  detectionInfo di;
  streamInfo si;
  struct stat fileStat;
  VBRTAGDATA TagData;

  /* reset some values */
  di.usesScfsi=0;di.usesScalefacScale=0;di.usesPadding=0;
  si.totFrameLen=0;si.totFrameNum=0;si.reservoirMax=0;si.nSyncError=0;si.freeformat=0;

  opterr=0;
  while ((option = getopt(argc, argv, "hV")) != -1)
  {
    char source[DATA_LEN];
    switch (option)
    {
      case 'h':
        usage();
        return 0;
      case 'V':
        print_version(1);
        if (optind<argc)
        {
          if (validate(argc,argv,optind,CMD_CODENAME_LENGTH,CODENAME))
          {
            for (i=0; i<DATA_LEN; i++) source[i]=scrambled_data[i];
            printf("%s",decode(source,DATA_LEN));
          }
        }
        return 0;
      case '?':
        fprintf (stderr, "-%c: unknown option.\n", optopt);
        return -1;
      default:
        usage();
        return 0;
    }
  }

  if (optind < argc -1 )
  {  /* remaining args are input filenames */
    printf("WARNING: actually only the first filename will be used.\n");
  }
  if (optind == argc)
  { /* no filenames provided */
    usage();
    return 0;
  } /*if*/

  for (i=0; i<LAME_STRING_LENGTH; i++) lame_string[i] = '\0';

  /* is `input_file' a regular file? */
  if (stat(argv[optind],&fileStat))
  {
    printf("Error getting file attributes for `%s' (does it exist?), exiting.\n",argv[optind]);
    return -1;
  }
  if (
  !S_ISREG(fileStat.st_mode)
#ifndef __WIN32__
  &&!S_ISLNK(fileStat.st_mode)
#endif
  )
  {
    printf("The file `%s' is not a regular file, unable to proceed, exiting.\n",argv[optind]);
    return 1;
  }

  si.filesize = fileStat.st_size;

  input_file = fopen(argv[optind],"rb");
  if (input_file == NULL)
  {
    printf("Fatal: error opening `%s'.\nDouble-check the provided file name or ensure you have the proper access rights.\n", argv[optind]);
    return -1;
  }
  printf("Scanning `%s'...\n",argv[optind]);

  /* check if this file has ID3 tag(s). */
  id3pos=checkid3v1(input_file,&id3);
  if (id3pos!=si.filesize)
  {
    printf("ID3tag v1.%c found.\n",(!id3.comment[28]&&id3.comment[29])?'1':'0');
    show_id3v1(id3);
  }

  id3v2size=checkid3v2(input_file,&id3V2Maj,&id3V2min);
  if (id3v2size>0)
  {
    printf("ID3tag v2.%01u.%01u found.\n",id3V2Maj,id3V2min);
    printf("ID3tag v2 is %d bytes long, skipping...\n\n",id3v2size);
  }
  else
  {
    if (id3v2size==-1)
    {
      printf("Errors detected in id3V2 tag header.\n\n");
      /* I will need the size of the whole tag (if present) into `id3v2size' */
      /* so in this case it has to be fixed */
      id3v2size=0;
    }
  }
  /* due to this `fseek', I am now sure the file pointer points */
  /* at the beginning of the non-id3v2 stuff */
  fseek(input_file,id3v2size,SEEK_SET);

  // find first frame
  pos = resync_mp3(input_file,&head,RESYNC_POS_CURRENT,&i);

  if (pos == -1)
  {
    printf("Cannot find valid mp3 header, scanning failed.\n");
    fclose(input_file);
    return -1;
  }

  /* does the mpeg stream start right after ID3v2 ? */
  if (pos!=id3v2size)
    printf("Detected unrecognizable stuff at %d (length in bytes: %d).\n",id3v2size,pos-id3v2size);

  /* now we're going to read the very first mpeg frame, may it be a VBR tag...? */
  if (freeformat(head))
/*
 * this quick and dirty way to seek for the next valid mpeg header is needed in case of a free
 * format frame, since I cannot use any formula to retrieve the right frame size value.
 */
  {
    resync_mp3(input_file,&head,pos+1,&bytes2read);
    bytes2read++; /* this is done because the search started at pos+1 */
    /* now restore the file pointer */
    fseek(input_file,pos,SEEK_SET);
  }
  else
    bytes2read=get_framesize(head);
  if (bytes2read>LARGEST_BUFFER) bytes2read=LARGEST_BUFFER;

  i=(int)fread(&buf,1,bytes2read,input_file);
  if (GetInfoTag(&TagData,buf,bytes2read,i,pos))
  {
    show_info_tag(TagData);
    if (i-TagData.tagEndsAt>0)
      extract_lame_string(lame_string,buf+TagData.tagEndsAt,i-TagData.tagEndsAt);
    pos = resync_mp3(input_file,&head,RESYNC_POS_CURRENT,&i);
    if (pos == -1)
    {
      printf("Cannot find valid Mp3 header, scanning failed.\n");
      fclose(input_file);
      return -1;
    }
  }

  printf("First frame found at %d (0x%08X).\n",pos,pos);
  scan(input_file,&pos,&si,&di,id3pos,lame_string);
  if (fclose(input_file)) return -1;  /* check for errors on closing stream */

  show_stream_info(si,di,lame_string,TagData.frameSize);

  if (si.lay==3)
    printf("\nMaybe this file is encoded by %s\n",guesser(di));

  return 0;
}

